/*
 * COL
 */

fn readStringN f n = (
	s = ""
	end = false
	for i = 1 to n do (
		c = readbyte f #unsigned
		if c == undefined do return undefined
		if c == 0 do end = true
		if not end do
			s += bit.intAsChar c
	)
	s
)

struct ColSphere (
	radius,
	pos,
	surf, piece,

	fn read f = (
		radius = readFloat f
		x = readFloat f
		y = readFloat f
		z = readFloat f
		pos = [x, y, z]
		surf = readByte f
		piece = readByte f
		fseek f 2 #seek_cur
	)
)

struct ColBox (
	v0, v1,
	surf, piece,

	fn read f = (
		x = readFloat f
		y = readFloat f
		z = readFloat f
		v0 = [x, y, z]
		x = readFloat f
		y = readFloat f
		z = readFloat f
		v1 = [x, y, z]
		surf = readByte f
		piece = readByte f
		fseek f 2 #seek_cur
	)
)

struct ColTriangle (
	v0, v1, v2,
	surf,

	fn read f = (
		v0 = readLong f
		v1 = readLong f
		v2 = readLong f
		surf = readByte f
		fseek f 3 #seek_cur
	)
)

struct assoc (
	key, val, file
)

fn keySort a1 a2 = (
	if a1.key < a2.key do return -1
	if a1.key > a2.key do return 1
	return 0
)

fn addToColIndex dict f filename = (
	--dict = #()
	while (fcc = readStringN f 4) == "COLL" do (
		len = readLong f
		name = readStringN f 24
		append dict (assoc key:name val:(ftell f) file:filename)
		fseek f (len-24) #seek_cur
	)
	--qsort dict keySort
	--dict
)

fn addToColIndexF dict filename = (
	f = fopen filename "rb"
	dict = addToColIndex dict f filename
	fclose f
)

-- break up mesh by material id
fn breakmesh m = (
	faces = #()
	for i = 1 to getNumFaces m do (
		id = getFaceMatID m i
		if faces[id] == undefined then
			faces[id] = #{i}
		else
			append faces[id] i
	)
	for f in faces collect (
		emesh = Editable_Mesh()
		emesh.mesh = meshop.detachFaces m f delete:false asMesh:true
		update emesh
		emesh
	)
)

fn createCol f = (
	-- bounding information
	fseek f 40 #seek_cur

	objs = #()

	numSpheres = readLong f
	for i = 1 to numSpheres do (
		sph = ColSphere()
		sph.read f

		colsph = sphere radius:sph.radius pos:sph.pos
		makeColObject colsph
		colsph.surf = sph.surf
		colsph.piece = sph.piece
		append objs colsph
	)

	-- unused, just ignore
	numLines = readLong f
	fseek f (numLines*24) #seek_cur

	numBoxes = readLong f
	for i = 1 to numBoxes do (
		cbox = ColBox()
		cbox.read f

		w = cbox.v1.x - cbox.v0.x
		l = cbox.v1.y - cbox.v0.y
		h = cbox.v1.z - cbox.v0.z
		x = (cbox.v1.x + cbox.v0.x)/2.0
		y = (cbox.v1.y + cbox.v0.y)/2.0
		colbx = box width:w length:l height:h pos:[x,y,cbox.v0.z]
		makeColObject colbx
		colbx.surf = cbox.surf
		colbx.piece = cbox.piece
		append objs colbx
	)

	verts = #()
	numVertices = readLong f
	for i = 1 to numVertices do (
		x = readFloat f
		y = readFloat f
		z = readFloat f
		append verts [x, y, z]
	)

	numTriangles = readLong f
	tris = #()
	matids = #()
	mats = #()
	fn getColMat mats m = (
		id = findItem mats m
		if id > 0 do return id
		append mats m
		return mats.count
	)
	for i = 1 to numTriangles do (
		tri = ColTriangle()
		tri.read f
		append tris [tri.v0+1, tri.v2+1, tri.v1+1]
		append matids (getColMat mats tri.surf)
	)
	if numTriangles != 0 do (
		colmesh = mesh vertices:verts faces:tris materialIDs:matids
		for i = 1 to numTriangles do
			setFaceSmoothGroup colmesh i 0
		m = breakmesh colmesh
		for i = 1 to mats.count do (
			makeColObject m[i]
			m[i].surf = mats[i]
			m[i].piece = 0
			append objs m[i]
		)
		delete colmesh
	)
	objs
)

fn findCol dict name = (
	for c in dict do
		if c.key == name do
			return c
	undefined
)

fn createColF filename offset = (
	f = fopen filename "rb"
	fseek f offset #seek_set
	objs = createCol f
	fclose f
	objs
)

coldict = #()
-- addToColIndexF coldict "C:/Users/aap/Desktop/out.col"
-- addToColIndexF coldict "C:\\Users\\aap\\Desktop\\gta3d\\data\\maps\\industsw\\industsw_col\\Chinabuilds08.col"
-- addToColIndexF coldict "C:\\Users\\aap\\Desktop\\gta3d\\data\\maps\\industsw\\industsw_col\\Chinabuilds05.col"
-- addToColIndexF coldict "C:\\Users\\aap\\Desktop\\gta3d\\data\\maps\\industsw\\industsw_col\\Chinabuilds07.col"
-- addToColIndexF coldict "C:\\Users\\aap\\Desktop\\3map\\industsw\\industsw_col\\Chinabuilds21.col"
-- addToColIndexF coldict "C:\\Users\\aap\\Desktop\\3map\\industsw\\industsw_col\\ind_land070.col"
-- addToColIndexF coldict "C:\\Users\\aap\\Desktop\\gta3d\\data\\maps\\industse\\industse_col\\ind_land124.col"
--addToColIndexF coldict "C:\\Users\\aap\\Desktop\\gta3d\\data\\maps\\industse\\industse_col\\ind_tendragblk.col"

--addToColIndexF coldict "C:\\Users\\aap\\games\\gta3\\data\\maps\\comsw\\comsw_col\\subentrance.col"
--addToColIndexF coldict "C:\\Users\\aap\\games\\gta3\\data\\maps\\comsw\\comsw_col\\subntrnce_b.col"
--addToColIndexF coldict "C:\\Users\\aap\\games\\gta3\\data\\maps\\industse\\industse_col\\ind_tendragblk.col"
addToColIndexF coldict "C:\\Users\\aap\\games\\gta3\\data\\maps\\industse\\industse_col\\block_maindrag.col"

c = coldict[1]
objs = createColF c.file c.val